#include "stdafx.h"
#include "tsip.h"
#include "mainwnd.h"

/* Methods defined in this file */
static DWORD WINAPI SerialThread(LPVOID);
static void ProcessPacket(BYTE* buffer, DWORD len);
static void ProcessPrimaryPacket(PRIMARY_TIMING_PACKET* packet);
static void ProcessSupplimentalPacket(SUPPLIMENTAL_TIMING_PACKET* packet);
static void SyncSystemTime(PRIMARY_TIMING_PACKET* packet);
static BOOL CheckLength(int expected, int actual);
void LogWarning(LPCTSTR fmtString, ...);

/* Constants */
#define DAYS_PER_WEEK 7
#define SECS_PER_DAY 86400

/* Beginning of GPS cycle 0 as modified Julian date */
#define GPS_EPOCH 44244

/* TSIP protocol characters */
#define _DLE ((BYTE)0x10)
#define _ETX ((BYTE)0x03)

/* Global/static variables */
static DWORD threadId = 0;
static HANDLE hThread = NULL;
static HANDLE hPort = NULL;
static HWND hWndMain = NULL;
static BOOL shutdownFlag = FALSE;
TCHAR errMsg[128];

/* Configurables */
LPCTSTR serialPortName = _T("COM1:");
UINT8  gpsCycle = 2;

/* Converts to/from big-endian */
inline void swap(UINT16* s)
{
#ifndef _BIG_ENDIAN
    BYTE tmp;
    BYTE* buf = (BYTE*)s;

    tmp = buf[0];
    buf[0] = buf[1];
    buf[1] = tmp;
#endif
}

/* Converts to/from big-endian */
inline void swap(INT16* s)
{
    swap((UINT16*)s);
}

/* Converts to/from big-endian */
inline void swap(UINT32* l)
{
#ifndef _BIG_ENDIAN
    BYTE tmp;
    BYTE* buf = (BYTE*)l;

    tmp = buf[0];
    buf[0] = buf[3];
    buf[3] = tmp;

    tmp = buf[1];
    buf[1] = buf[2];
    buf[2] = tmp;
#endif
}

/* Converts to/from big-endian */
inline void swap(INT32* s)
{
    swap((UINT32*)s);
}

/* Converts to/from big-endian */
inline void swap(float* f)
{
    swap((UINT32*)f);
}

/* Converts to/from big-endian */
inline void swap(double* d)
{
#ifndef _BIG_ENDIAN
    BYTE tmp;
    BYTE* buf = (BYTE*)d;

    tmp = buf[0];
    buf[0] = buf[7];
    buf[7] = tmp;

    tmp = buf[1];
    buf[1] = buf[6];
    buf[6] = tmp;

    tmp = buf[2];
    buf[2] = buf[5];
    buf[5] = tmp;
    
    tmp = buf[3];
    buf[3] = buf[4];
    buf[4] = tmp;
#endif
}

/*
 * Convert Modified Julian Day to calendar date.
 * - Assumes Gregorian calendar.
 * - Adapted from Fliegel/van Flandern ACM 11/#10 p 657 Oct 1968.
 * (from http://www.leapsecond.com/tools/gpsdate.c)
 */

void MjdToDate(long mjd, UINT16 *year, UINT8 *month, UINT8 *day)
{
    long J, C, Y, M;

    J = mjd + 2400001 + 68569;
    C = 4 * J / 146097;
    J = J - (146097 * C + 3) / 4;
    Y = 4000 * (J + 1) / 1461001;
    J = J - 1461 * Y / 4 + 31;
    M = 80 * J / 2447;
    *day = J - 2447 * M / 80;
    J = M / 11;
    *month = (UINT8)(M + 2 - (12 * J));
    *year = (UINT16)(100 * (C - 49) + Y + J);
}

/* Called by the main thread to open the serial port
   and start a background thread 
*/
BOOL StartSerialThread(HWND main)
{
    DCB PortDCB;
    COMMTIMEOUTS CommTimeouts;
    DWORD dwError;
        
    hWndMain = main;
    hPort = CreateFile(serialPortName, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, NULL);
    
    // If it fails to open the port, return FALSE.
    if ( hPort == INVALID_HANDLE_VALUE ) 
    {
        dwError = GetLastError ();
        MessageBox(hWndMain, _T("Could not open serial port"), serialPortName, MB_OK | MB_ICONERROR);
        return FALSE;
    }
    
    PortDCB.DCBlength = sizeof (DCB);     
    
    // Get the default port setting information.
    GetCommState (hPort, &PortDCB);
    
    // Change the DCB structure settings.
    PortDCB.BaudRate = 9600;              // Current baud 
    PortDCB.fBinary = TRUE;               // Binary mode; no EOF check 
    PortDCB.fParity = TRUE;               // Enable parity checking. 
    PortDCB.fOutxCtsFlow = FALSE;         // No CTS output flow control 
    PortDCB.fOutxDsrFlow = FALSE;         // No DSR output flow control 
    PortDCB.fDtrControl = DTR_CONTROL_ENABLE; 
    // DTR flow control type 
    PortDCB.fDsrSensitivity = FALSE;      // DSR sensitivity 
    PortDCB.fTXContinueOnXoff = TRUE;     // XOFF continues Tx 
    PortDCB.fOutX = FALSE;                // No XON/XOFF out flow control 
    PortDCB.fInX = FALSE;                 // No XON/XOFF in flow control 
    PortDCB.fErrorChar = FALSE;           // Disable error replacement. 
    PortDCB.fNull = FALSE;                // Disable null stripping. 
    PortDCB.fRtsControl = RTS_CONTROL_ENABLE; 
    // RTS flow control 
    PortDCB.fAbortOnError = FALSE;        // Do not abort reads/writes on 
    // error.
    PortDCB.ByteSize = 8;                 // Number of bits/bytes, 4-8 
    PortDCB.Parity = NOPARITY;            // 0-4=no,odd,even,mark,space 
    PortDCB.StopBits = ONESTOPBIT;        // 0,1,2 = 1, 1.5, 2 
    
    // Configure the port according to the specifications of the DCB 
    // structure.
    if (!SetCommState (hPort, &PortDCB))
    {
        dwError = GetLastError();
        MessageBox (hWndMain, _T("Unable to configure the serial port"), 
            _T("Error"), MB_OK);

        CloseHandle(hPort);
        hPort = NULL;
        return FALSE;
    }
    
    // Retrieve the time-out parameters for all read and write operations
    // on the port. 
    GetCommTimeouts (hPort, &CommTimeouts);
    
    // Change the COMMTIMEOUTS structure settings.
    CommTimeouts.ReadIntervalTimeout = 10;  
    CommTimeouts.ReadTotalTimeoutMultiplier = 0;  
    CommTimeouts.ReadTotalTimeoutConstant = 2000; 
    CommTimeouts.WriteTotalTimeoutMultiplier = 10;  
    CommTimeouts.WriteTotalTimeoutConstant = 1000;    
    
    // Set the time-out parameters for all read and write operations
    // on the port. 
    if (!SetCommTimeouts (hPort, &CommTimeouts))
    {
        dwError = GetLastError();
        MessageBox (hWndMain, TEXT("Unable to set the time-out parameters"), 
            TEXT("Error"), MB_OK);

        CloseHandle(hPort);
        hPort = NULL;

        return FALSE;
    }
    
    // Direct the port to perform extended functions SETDTR and SETRTS.
    // SETDTR: Sends the DTR (data-terminal-ready) signal.
    // SETRTS: Sends the RTS (request-to-send) signal. 
    EscapeCommFunction (hPort, SETDTR);
    EscapeCommFunction (hPort, SETRTS);
    
    // Create a read thread for reading data from the communication port.
    hThread = CreateThread (NULL, 0, SerialThread, hWndMain, 0, &threadId);
    if (hThread == INVALID_HANDLE_VALUE)
    {
        dwError = GetLastError();
        MessageBox (hWndMain, TEXT("Unable to create the read thread"), 
            TEXT("Error"), MB_OK);

        CloseHandle(hPort);
        hPort = NULL;

        return FALSE;
    }
    
    return TRUE;
}

/* Main function for the background thread
 */
DWORD WINAPI SerialThread(LPVOID)
{
    BYTE serialBuffer[256];
    DWORD remaining;

    BYTE packet[256];
    BYTE* packetTail = packet;
 
    int state = 0;

    while (!shutdownFlag)
    {
        if (ReadFile(hPort, serialBuffer, sizeof(serialBuffer), &remaining, NULL))
        {
            BYTE* c = serialBuffer;

            if (remaining == 0)
            {
                // Timeout occurred.  Reset state machine
                state = 0;
                packetTail = packet;
                SendMessage(hWndMain, WM_NO_CONTACT, 0, 0);
            }

            /* Parse incoming bytes using a state machine to detect packet boundaries */
            while (remaining > 0)
            {
                BOOL ignore = TRUE;

                switch (state)
                {
                case 0:
                    switch (*c)
                    {
                        case _DLE:
                            state = 1;
                            break;

                        default:
                            break;
                    }
                    break;

                case 1:
                    switch (*c)
                    {
                    case _DLE:
                        state = 2;
                        break;

                    default:
                        ignore = FALSE;
                        break;
                    }
                    break;

                case 2:
                    switch (*c)
                    {
                    case _DLE:
                        state = 1;
                        ignore = FALSE;
                        break;

                    case _ETX:
                        state = 0;
                        ProcessPacket(packet, packetTail - packet);
                        packetTail = packet;
                        break;

                    default:
                        // We must have jumped in mid-packet, restart packet
                        state = 1;
                        packetTail = packet;
                        ignore = FALSE;
                        break;
                    }
                    break;
                }

                if (!ignore)
                {
                    // Append character to packet buffer
                    *packetTail = *c;
                    packetTail++;

                    if (packetTail > packet + sizeof(packet))
                    {
                        wsprintf(errMsg,_T("ERROR: packet exceeded buffer length\n"));
                        OutputDebugString(errMsg);
                        state = 0;
                        packetTail = packet;
                    }
                }

                c++;
                remaining--;
            }
        }
    }

    CloseHandle(hPort);
    hPort = NULL;
    return 0;
}

BOOL StopSerialThread()
{
    if (hPort != NULL)
    {
        CloseHandle(hPort);
    }

    shutdownFlag = TRUE;
    return TRUE;
}

void ProcessPacket(BYTE* buffer, DWORD len)
{
    if (len < 2)
    {
        // Packet too short
        wsprintf(errMsg, _T("ProcessPacket: Unable to read packet type, only %d bytes received"), len);
        OutputDebugString(errMsg);
    }
    else
    {
        switch (buffer[0])
        {
        case 0x8f:
            switch (buffer[1])
            {
            case 0xAB:
                if (CheckLength(sizeof(PRIMARY_TIMING_PACKET), len))
                {
                    ProcessPrimaryPacket((PRIMARY_TIMING_PACKET*)buffer);
                }
                break;
            case 0xAC:
                if (CheckLength(sizeof(SUPPLIMENTAL_TIMING_PACKET), len))
                {
                    ProcessSupplimentalPacket((SUPPLIMENTAL_TIMING_PACKET*)buffer);
                }
                break;
            default:
                wsprintf(errMsg, _T("ProcessPacket: Ignoring packet 0x8f-%2x (length %d)\n"), (int)buffer[1], len);
                OutputDebugString(errMsg);
                break;
            }
            break;

        default:
            wsprintf(errMsg, _T("ProcessPacket: Ignoring packet 0x%2x (length %d)\n"), (int)buffer[0], len);
            OutputDebugString(errMsg);
            break; 
        }
    }
}

/* Work around GPS week rollover bug in Thunderbolt */
void FixDate(struct PRIMARY_TIMING_PACKET* packet) 
{
    // Remove Trimble's hi bit hack and add in real cycle
    packet->weekNumber = (packet->weekNumber & 1023) | (gpsCycle << 10);

    // Recompute month, day, year
	long dayOfWeek = packet->timeOfWeek / SECS_PER_DAY;
    long timeOfDay = packet->timeOfWeek % SECS_PER_DAY;
    long mjd = GPS_EPOCH + (packet->weekNumber * DAYS_PER_WEEK) + dayOfWeek;

	// Is it still yesterday UTC?
    if (packet->flags & FLAG_UTC && timeOfDay < packet->utcOffset) 
	{
        mjd--;
    }

    MjdToDate(mjd, &packet->year, &packet->month, &packet->dayOfMonth);
}

void ProcessPrimaryPacket(struct PRIMARY_TIMING_PACKET* packet)
{
    swap(&packet->timeOfWeek);
    swap(&packet->weekNumber);
    swap(&packet->utcOffset);
    swap(&packet->year);

    FixDate(packet);

    if (packet->seconds == 10)
    {
        SyncSystemTime(packet);
    }
/*
    wsprintf(errMsg, _T("Primary: %02hhd:%02hhd:%02hhd %02hhd-%02hhd-%04hd flags=%02hhx\n"),
        packet->hours, packet->minutes, packet->seconds,
        packet->month, packet->dayOfMonth, packet->year, packet->flags);
    OutputDebugString(errMsg);*/

    SendMessage(hWndMain, WM_PRIMARY_PACKET, (WPARAM)packet, NULL);
}


void ProcessSupplimentalPacket(struct SUPPLIMENTAL_TIMING_PACKET* packet)
{
    swap(&packet->holdoverDuration);
    swap(&packet->criticalAlarms);
    swap(&packet->minorAlarms);
    swap(&packet->ppsOffset);
    swap(&packet->tenMhzOffset);
    swap(&packet->dacValue);
    swap(&packet->dacVoltage);
    swap(&packet->temperature);
    swap(&packet->latitude);
    swap(&packet->longitude);
    swap(&packet->altitude);

/*    wprintf(_T("Secondary: %s critical=%04hx minor=%04hx pps=%f 10mhz=%f\n"),
        LookupString(packet->discipliningMode, discipliningModes, sizeof(discipliningModes)/sizeof(LPCTSTR)),
        packet->criticalAlarms, packet->minorAlarms,
        packet->ppsOffset, packet->tenMhzOffset);*/

    SendMessage(hWndMain, WM_SUPPLIMENTAL_PACKET, (WPARAM)packet, NULL);
}

// Adjust system time if necessary
void SyncSystemTime(PRIMARY_TIMING_PACKET* packet)
{
    SYSTEMTIME systemTime;

    GetSystemTime(&systemTime);
    if (systemTime.wMinute != packet->minutes ||
        systemTime.wSecond < packet->seconds -1 || 
        systemTime.wSecond > packet->seconds +1 ||
        systemTime.wHour != packet->hours ||
        systemTime.wDay != packet->dayOfMonth ||
        systemTime.wMonth != packet->month ||
        systemTime.wYear != packet->year)
    {
        systemTime.wYear = packet->year;
        systemTime.wMonth = packet->month;
        systemTime.wDay = packet->dayOfMonth;
        systemTime.wHour = packet->hours;
        systemTime.wMinute = packet->minutes;
        systemTime.wSecond = packet->seconds;
        systemTime.wMilliseconds = 100; // Just a guess
        SetSystemTime(&systemTime);
    }
}

BOOL CheckLength(int expected, int actual)
{
    if (expected != actual)
    {
        wsprintf(errMsg, _T("Error: incorrect packet length.  Expected %d, got %d\n"), expected, actual);
        OutputDebugString(errMsg);
        return FALSE;
    }
    else

    {
        return TRUE;
    }
}

